==> matrix_sdk-0.14.0.tar.gz: lib/src/user.dart <==
// Copyright (C) 2019  Wilko Manger
// Copyright (C) 2019  Mathieu Velten
//
// This file is part of matrix-dart-sdk.
//
// matrix-dart-sdk is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// matrix-dart-sdk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with matrix-dart-sdk.  If not, see <https://www.gnu.org/licenses/>.

import 'dart:convert';

import 'package:chopper/chopper.dart';
import 'package:kiwi/kiwi.dart';
import 'package:matrix_sdk/src/api/access_token_interceptor.dart';
import 'package:matrix_sdk/src/api/client.dart';
import 'package:matrix_sdk/src/event/room/room_event.dart';
import 'package:matrix_sdk/src/homeserver.dart';
import 'package:matrix_sdk/src/identifier.dart';
import 'package:matrix_sdk/src/room/room.dart';
import 'package:matrix_sdk/src/room/rooms.dart';
import 'package:matrix_sdk/src/store.dart';
import 'package:matrix_sdk/src/util/chopper.dart';
import 'package:meta/meta.dart';

class UserState implements Comparable<UserState> {
  final RoomId roomId;
  final DateTime since;

  final String displayName;
  final Uri avatarUrl;

  UserState({
    @required this.roomId,
    @required this.since,
    this.displayName,
    this.avatarUrl
  });

  @override
  int compareTo(UserState other) => since.compareTo(other.since);

  @override
  bool operator ==(dynamic other) {
    if (other is UserState) {
      return this.roomId == other.roomId
        && this.since == other.since
        && this.displayName == other.displayName
        && this.avatarUrl == other.avatarUrl;
    }

    return false;
  }

  @override
  int get hashCode
    => roomId.hashCode
     + since.hashCode
     + displayName.hashCode
     + avatarUrl.hashCode;
}

/// A User with at a specific [state].
class User extends Identifiable<UserId> implements Comparable<User> {
  @override
  final UserId id;

  final UserState state;

  User({@required this.id, @required this.state});

  String get name => state?.displayName;
  Uri get avatarUrl => state?.avatarUrl;

  /// Returns true when this user is the same
  /// as [user]. Identical to `this.id == user.id`
  bool isIdenticalTo(User user) => this.id == user.id;

  @override
  int compareTo(User other) => state.compareTo(other.state);

  @override
  bool operator ==(dynamic other) {

    if (other is User) {
      return this.id == other.id
        && this.state == other.state;
    }

    return false;
  }

  @override
  int get hashCode => id.hashCode + state.hashCode;
}

/// A user which is authenticated and can send messages, join rooms etc.
class LocalUser extends User {
  final Store _store;

  final Homeserver homeserver;
  final String accessToken;

  Rooms _rooms;
  Rooms get rooms => _rooms;

  bool _isLoggedOut = false;
  bool get isLoggedOut => _isLoggedOut;

  /// @internal
  final String scope;

  String _syncToken;

  final ChopperClient _chopper;
  ClientService get _client => _chopper.getService();

  static Future<LocalUser> fromStore(Store store) async {
    final isOpen = await store.open();

    if (!isOpen) {
      return null;
    }

    final userId = await store.getUserId();
    if (userId == null) {
      return null;
    }

    final hs = await store.getHomeserver();
    if (hs == null) {
      return null;
    }

    final accessToken = await store.getAccessToken();
    if (accessToken == null) {
      return null;
    }

    return LocalUser(
      id: userId,
      homeserver: hs,
      accessToken: accessToken,
      store: store
    );
  }

  LocalUser(
    {@required UserId id,
    @required this.homeserver,
    @required this.accessToken,
    @required Store store})
    : _chopper = createChopper(
          url: homeserver.url,
          interceptors: [AccessTokenInterceptor(accessToken)]),
      _store = store,
      scope = homeserver.scope + id.toString(),
      super(id: id) {
    Container().registerInstance(this, name: scope);
    Container().registerInstance(_client, name: scope);
    Container().registerInstance(store, name: scope);

    _rooms = Rooms(store: store, scope: scope);
  }

  /// Invalidates the access token of the user. Makes all
  /// [LocalUser] calls unusable.
  Future<void> logout() async {
    await _client.logout();
    _isLoggedOut = true;
  }

  /// Send all unsent messages still in the [Store].
  Future<void> sendAllUnsent() async {
    await for (Room room in rooms.all()) {
      if (room is JoinedRoom) {
        await for (RoomEvent event in _store.getUnsentEvents(room)) {
          await for (final sentState in room.send(
            event.content,
            transactionId: event.unsigned?.transactionId
          )) { }
        }
      }
    }
  }

  Future<bool> setPusher(String pushKey, url, appId, appName, deviceName,
      {lang = 'en-US', remove = false, eventIdOnly = false}) async {

    var data = {
      'lang': lang,
      'kind': 'http',
      'data': {'url': url},
      'device_display_name': deviceName,
      'app_id': appId,
      'app_display_name': appName,
      'pushkey': pushKey,
    };
    if (remove) {
      data['kind'] = 'null';
    }
    if (eventIdOnly) {
      data['data']['format'] = 'event_id_only';
    }

    final response = await _client.setPusher(json.encode(data));

    switch (response.statusCode) {
      case 200:
        return true;
      default:
        return false;
    }
  }

  // Sync //
  /// Syncs data with the user's [homeserver]. Use the returned [Stream] to
  /// get when the sync has been successful (`true`) or unsuccessful (`false`).
  Stream<SyncState> sync(
    {Duration maxRetryAfter = const Duration(seconds: 30)}) async* {

    if (isLoggedOut) {
      throw StateError('The user can not be logged out');
    }

    // This var is used to implements exponential backoff
    // until it reaches maxRetryAfter
    var retryAfter = 500;

    while (true) {
      final state = await syncOnce(timeout: Duration(seconds: 10));
      yield state;
      if (!state.succeeded) {
        await Future.delayed(Duration(milliseconds: retryAfter));
        retryAfter = (retryAfter * 1.5).floor();
        if (retryAfter > maxRetryAfter.inMilliseconds) {
          retryAfter = maxRetryAfter.inMilliseconds;
        }
      } else {
        // reset exponential backoff
        retryAfter = 500;
      }
    }
  }

  Future<SyncState> syncOnce({bool updateSyncToken = true, Map<String, dynamic> filter,
      timeout = Duration.zero, bool fullState = false}) async {

    if (isLoggedOut) {
      throw StateError('The user can not be logged out');
    }

    _syncToken = await _store?.getSyncToken() ?? _syncToken;

    if (filter == null) { filter = Map<String, dynamic>(); }
    if (filter['room'] == null) { filter['room'] = Map<String, dynamic>(); }
    if (filter['room']['state'] == null) {filter['room']['state'] = Map<String, dynamic>(); }

    filter['room']['state']['lazy_load_members'] = true;

    try {
      Response response = await _client.sync(
          since: _syncToken,
          fullState: fullState,
          filter: json.encode(filter),
          timeout: timeout.inMilliseconds);
      final body = json.decode(response.body);

      await _processRooms(body, body['account_data']);

      if (updateSyncToken) {
        // Store sync token
        _syncToken = body['next_batch'] as String;
        await _store?.setSyncToken(_syncToken);
      }

      return SyncState();
    } on Exception catch (e) {
      return SyncState(
        exception: e
      );
    }
  }

  Future<void> _processRooms(
      Map<String, dynamic> body,
      Map<String, dynamic> accountData) async {
    final jRooms = body['rooms'];

    final processed = List<Room>();

    const join = 'join';
    const leave = 'leave';

    process(Map<String, dynamic> jsonRooms, {@required String type}) async {
      if (jsonRooms != null) {
        for (final entry in jsonRooms.entries) {
          final roomId = RoomId(entry.key);
          final json = entry.value;

          Room room = await rooms[roomId];
          if (room == null) {
            switch(type) {
              case join:
                room = JoinedRoom(id: roomId, scope: scope);
                break;
              case leave:
                room = LeftRoom(id: roomId, scope: scope);;
                break;
            }
          } else {
            if (type == leave) {
              room = LeftRoom(
                id: room.id,
                scope: scope,
                name: room.name,
                avatarUrl: room.avatarUrl,
                directUser: room.directUser,
                creator: room.creator,
                previousBatch: room.previousBatch
              );
            }
          }

          await room.processSync(json);
          await room.processUserAccountData(accountData);

          processed.add(room);
        }
      }
    }

    await process(jRooms[join], type: join);
    await process(jRooms[leave], type: leave);

    // Save rooms
    await rooms.addAll(processed);
  }

  // End Sync //
}

class SyncState {
  final bool succeeded;

  /// If [succeeded] is false, this is the exception that was thrown.
  final Exception exception;

  SyncState({this.exception}) : succeeded = exception == null;
}

==> matrix_sdk-0.16.0.tar.gz: lib/src/user.dart <==
// Copyright (C) 2019  Wilko Manger
// Copyright (C) 2019  Mathieu Velten
//
// This file is part of matrix-dart-sdk.
//
// matrix-dart-sdk is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// matrix-dart-sdk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with matrix-dart-sdk.  If not, see <https://www.gnu.org/licenses/>.

import 'dart:convert';

import 'package:chopper/chopper.dart';
import 'package:kiwi/kiwi.dart';
import 'package:matrix_sdk/src/api/access_token_interceptor.dart';
import 'package:matrix_sdk/src/api/client.dart';
import 'package:matrix_sdk/src/event/room/room_event.dart';
import 'package:matrix_sdk/src/homeserver.dart';
import 'package:matrix_sdk/src/identifier.dart';
import 'package:matrix_sdk/src/room/room.dart';
import 'package:matrix_sdk/src/room/rooms.dart';
import 'package:matrix_sdk/src/store.dart';
import 'package:matrix_sdk/src/util/chopper.dart';
import 'package:meta/meta.dart';
import 'package:url/url.dart';

class UserState implements Comparable<UserState> {
  final RoomId roomId;
  final DateTime since;

  final String displayName;
  final Url avatarUrl;

  UserState({
    @required this.roomId,
    @required this.since,
    this.displayName,
    this.avatarUrl
  });

  @override
  int compareTo(UserState other) => since.compareTo(other.since);

  @override
  bool operator ==(dynamic other) {
    if (other is UserState) {
      return this.roomId == other.roomId
        && this.since == other.since
        && this.displayName == other.displayName
        && this.avatarUrl == other.avatarUrl;
    }

    return false;
  }

  @override
  int get hashCode
    => roomId.hashCode
     + since.hashCode
     + displayName.hashCode
     + avatarUrl.hashCode;
}

/// A User with at a specific [state].
class User extends Identifiable<UserId> implements Comparable<User> {
  @override
  final UserId id;

  final UserState state;

  User({@required this.id, @required this.state});

  String get name => state?.displayName;
  Url get avatarUrl => state?.avatarUrl;

  /// Returns true when this user is the same
  /// as [user]. Identical to `this.id == user.id`
  bool isIdenticalTo(User user) => this.id == user.id;

  @override
  int compareTo(User other) => state.compareTo(other.state);

  @override
  bool operator ==(dynamic other) {

    if (other is User) {
      return this.id == other.id
        && this.state == other.state;
    }

    return false;
  }

  @override
  int get hashCode => id.hashCode + state.hashCode;
}

/// A user which is authenticated and can send messages, join rooms etc.
class LocalUser extends User {
  final Store _store;

  final Homeserver homeserver;
  final String accessToken;

  Rooms _rooms;
  Rooms get rooms => _rooms;

  bool _isLoggedOut = false;
  bool get isLoggedOut => _isLoggedOut;

  /// @internal
  final String scope;

  String _syncToken;

  final ChopperClient _chopper;
  ClientService get _client => _chopper.getService();

  static Future<LocalUser> fromStore(Store store) async {
    final isOpen = await store.open();

    if (!isOpen) {
      return null;
    }

    final userId = await store.getUserId();
    if (userId == null) {
      return null;
    }

    final hs = await store.getHomeserver();
    if (hs == null) {
      return null;
    }

    final accessToken = await store.getAccessToken();
    if (accessToken == null) {
      return null;
    }

    return LocalUser(
      id: userId,
      homeserver: hs,
      accessToken: accessToken,
      store: store
    );
  }

  LocalUser(
    {@required UserId id,
    @required this.homeserver,
    @required this.accessToken,
    @required Store store})
    : _chopper = createChopper(
          url: homeserver.url,
          interceptors: [AccessTokenInterceptor(accessToken)]),
      _store = store,
      scope = homeserver.scope + id.toString(),
      super(id: id) {
    Container().registerInstance(this, name: scope);
    Container().registerInstance(_client, name: scope);
    Container().registerInstance(store, name: scope);

    _rooms = Rooms(store: store, scope: scope);
  }

  /// Invalidates the access token of the user. Makes all
  /// [LocalUser] calls unusable.
  Future<void> logout() async {
    await _client.logout();
    _isLoggedOut = true;
  }

  /// Send all unsent messages still in the [Store].
  Future<void> sendAllUnsent() async {
    await for (Room room in rooms.get()) {
      if (room is JoinedRoom) {
        await for (RoomEvent event in _store.getUnsentEvents(room)) {
          await for (final sentState in room.send(
            event.content,
            transactionId: event.unsigned?.transactionId
          )) { }
        }
      }
    }
  }

  Future<bool> setPusher(String pushKey, url, appId, appName, deviceName,
      {lang = 'en-US', remove = false, eventIdOnly = false}) async {

    var data = {
      'lang': lang,
      'kind': 'http',
      'data': {'url': url},
      'device_display_name': deviceName,
      'app_id': appId,
      'app_display_name': appName,
      'pushkey': pushKey,
    };
    if (remove) {
      data['kind'] = 'null';
    }
    if (eventIdOnly) {
      data['data']['format'] = 'event_id_only';
    }

    final response = await _client.setPusher(json.encode(data));

    switch (response.statusCode) {
      case 200:
        return true;
      default:
        return false;
    }
  }

  // Sync //
  /// Syncs data with the user's [homeserver]. Use the returned [Stream] to
  /// get when the sync has been successful (`true`) or unsuccessful (`false`).
  Stream<SyncState> sync(
    {Duration maxRetryAfter = const Duration(seconds: 30)}) async* {

    if (isLoggedOut) {
      throw StateError('The user can not be logged out');
    }

    // This var is used to implements exponential backoff
    // until it reaches maxRetryAfter
    var retryAfter = 500;

    while (true) {
      final state = await syncOnce(timeout: Duration(seconds: 10));
      yield state;
      if (!state.succeeded) {
        await Future.delayed(Duration(milliseconds: retryAfter));
        retryAfter = (retryAfter * 1.5).floor();
        if (retryAfter > maxRetryAfter.inMilliseconds) {
          retryAfter = maxRetryAfter.inMilliseconds;
        }
      } else {
        // reset exponential backoff
        retryAfter = 500;
      }
    }
  }

  Future<SyncState> syncOnce({bool updateSyncToken = true, Map<String, dynamic> filter,
      timeout = Duration.zero, bool fullState = false}) async {

    if (isLoggedOut) {
      throw StateError('The user can not be logged out');
    }

    _syncToken = await _store?.getSyncToken() ?? _syncToken;

    if (filter == null) { filter = Map<String, dynamic>(); }
    if (filter['room'] == null) { filter['room'] = Map<String, dynamic>(); }
    if (filter['room']['state'] == null) {filter['room']['state'] = Map<String, dynamic>(); }

    filter['room']['state']['lazy_load_members'] = true;

    try {
      Response response = await _client.sync(
          since: _syncToken,
          fullState: fullState,
          filter: json.encode(filter),
          timeout: timeout.inMilliseconds);
      final body = json.decode(response.body);

      await _processRooms(body, body['account_data']);

      if (updateSyncToken) {
        // Store sync token
        _syncToken = body['next_batch'] as String;
        await _store?.setSyncToken(_syncToken);
      }

      return SyncState();
    } on Exception catch (e) {
      return SyncState(
        exception: e
      );
    }
  }

  Future<void> _processRooms(
      Map<String, dynamic> body,
      Map<String, dynamic> accountData) async {
    final jRooms = body['rooms'];

    final processed = List<Room>();

    const join = 'join';
    const leave = 'leave';

    process(Map<String, dynamic> jsonRooms, {@required String type}) async {
      if (jsonRooms != null) {
        for (final entry in jsonRooms.entries) {
          final roomId = RoomId(entry.key);
          final json = entry.value;

          Room room = await rooms[roomId];
          if (room == null) {
            switch(type) {
              case join:
                room = JoinedRoom(id: roomId, scope: scope);
                break;
              case leave:
                room = LeftRoom(id: roomId, scope: scope);;
                break;
            }
          } else {
            if (type == leave) {
              room = LeftRoom(
                id: room.id,
                scope: scope,
                name: room.name,
                avatarUrl: room.avatarUrl,
                directUser: room.directUser,
                creator: room.creator,
                previousBatch: room.previousBatch
              );
            }
          }

          await room.processSync(json);
          await room.processUserAccountData(accountData);

          processed.add(room);
        }
      }
    }

    await process(jRooms[join], type: join);
    await process(jRooms[leave], type: leave);

    // Save rooms
    await rooms.addAll(processed);
  }

  // End Sync //
}

class SyncState {
  final bool succeeded;

  /// If [succeeded] is false, this is the exception that was thrown.
  final Exception exception;

  SyncState({this.exception}) : succeeded = exception == null;
}
